import filecmp
import os
import time
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest.mock import patch

import pytest
from truss.constants import (
    BASE_TRTLLM_REQUIREMENTS,
    OPENAI_COMPATIBLE_TAG,
    TRTLLM_BASE_IMAGE,
    TRTLLM_PREDICT_CONCURRENCY,
    TRTLLM_PYTHON_EXECUTABLE,
    TRTLLM_TRUSS_DIR,
)
from truss.contexts.image_builder.serving_image_builder import (
    HF_ACCESS_TOKEN_FILE_NAME,
    ServingImageBuilderContext,
    get_files_to_cache,
)
from truss.tests.test_testing_utilities_for_other_tests import ensure_kill_all
from truss.truss_config import ModelCache, ModelRepo, TrussConfig
from truss.truss_handle import TrussHandle

BASE_DIR = Path(__file__).parent


def test_serving_image_dockerfile_from_user_base_image(custom_model_truss_dir):
    th = TrussHandle(custom_model_truss_dir)
    th.set_base_image("baseten/truss-server-base:3.9-v0.4.3", "/usr/local/bin/python3")
    builder_context = ServingImageBuilderContext
    image_builder = builder_context.run(th.spec.truss_dir)
    with TemporaryDirectory() as tmp_dir:
        tmp_path = Path(tmp_dir)
        image_builder.prepare_image_build_dir(tmp_path)
        with open(tmp_path / "Dockerfile", "r") as f:
            gen_docker_lines = f.readlines()
        with open(
            f"{BASE_DIR}/../../../test_data/server.Dockerfile",
            "r",
        ) as f:
            server_docker_lines = f.readlines()

        def filter_empty_lines(lines):
            return list(filter(lambda x: x and x != "\n" and x != "", lines))

        gen_docker_lines = filter_empty_lines(gen_docker_lines)
        server_docker_lines = filter_empty_lines(server_docker_lines)
        assert gen_docker_lines == server_docker_lines


def test_requirements_setup_in_build_dir(custom_model_truss_dir):
    th = TrussHandle(custom_model_truss_dir)
    th.add_python_requirement("numpy")
    builder_context = ServingImageBuilderContext
    image_builder = builder_context.run(th.spec.truss_dir)

    with TemporaryDirectory() as tmp_dir:
        tmp_path = Path(tmp_dir)
        image_builder.prepare_image_build_dir(tmp_path)
        with open(tmp_path / "requirements.txt", "r") as f:
            requirements_content = f.read()

        with open(f"{BASE_DIR}/../../../templates/server/requirements.txt", "r") as f:
            base_requirements_content = f.read()

        assert requirements_content == base_requirements_content + "numpy\n"


def flatten_cached_files(local_cache_files):
    return [file.source for file in local_cache_files]


def test_correct_hf_files_accessed_for_caching():
    model = "openai/whisper-small"
    config = TrussConfig(
        python_version="py39",
        model_cache=ModelCache(models=[ModelRepo(repo_id=model)]),
    )

    with TemporaryDirectory() as tmp_dir:
        truss_path = Path(tmp_dir)
        build_path = truss_path / "build"
        build_path.mkdir(parents=True, exist_ok=True)

        hf_path = Path("root/.cache/huggingface/hub")

        model_files, files_to_cache = get_files_to_cache(config, truss_path, build_path)
        files_to_cache = flatten_cached_files(files_to_cache)
        assert str(hf_path / "version.txt") in files_to_cache

        blobs = [
            blob
            for blob in files_to_cache
            if blob.startswith(f"{hf_path}/models--openai--whisper-small/blobs/")
        ]
        assert len(blobs) >= 1

        files = model_files[model]["files"]
        assert "model.safetensors" in files
        assert "tokenizer_config.json" in files


@patch("truss.contexts.image_builder.serving_image_builder.GCSCache.list_files")
def test_correct_gcs_files_accessed_for_caching(mock_list_bucket_files):
    mock_list_bucket_files.return_value = [
        "fake_model-001-of-002.bin",
        "fake_model-002-of-002.bin",
    ]
    model = "gs://crazy-good-new-model-7b"

    config = TrussConfig(
        python_version="py39",
        model_cache=ModelCache(models=[ModelRepo(repo_id=model)]),
    )

    with TemporaryDirectory() as tmp_dir:
        truss_path = Path(tmp_dir)
        build_path = truss_path / "build"
        build_path.mkdir(parents=True, exist_ok=True)

        model_files, files_to_cache = get_files_to_cache(config, truss_path, build_path)
        files_to_cache = flatten_cached_files(files_to_cache)

        assert (
            "/app/model_cache/crazy-good-new-model-7b/fake_model-001-of-002.bin"
            in files_to_cache
        )
        assert (
            "/app/model_cache/crazy-good-new-model-7b/fake_model-002-of-002.bin"
            in files_to_cache
        )

        assert "fake_model-001-of-002.bin" in model_files[model]["files"]
        assert "fake_model-001-of-002.bin" in model_files[model]["files"]


@patch("truss.contexts.image_builder.serving_image_builder.S3Cache.list_files")
def test_correct_s3_files_accessed_for_caching(mock_list_bucket_files):
    mock_list_bucket_files.return_value = [
        "fake_model-001-of-002.bin",
        "fake_model-002-of-002.bin",
    ]
    model = "s3://crazy-good-new-model-7b"

    config = TrussConfig(
        python_version="py39",
        model_cache=ModelCache(models=[ModelRepo(repo_id=model)]),
    )

    with TemporaryDirectory() as tmp_dir:
        truss_path = Path(tmp_dir)
        build_path = truss_path / "build"
        build_path.mkdir(parents=True, exist_ok=True)

        model_files, files_to_cache = get_files_to_cache(config, truss_path, build_path)
        files_to_cache = flatten_cached_files(files_to_cache)

        assert (
            "/app/model_cache/crazy-good-new-model-7b/fake_model-001-of-002.bin"
            in files_to_cache
        )
        assert (
            "/app/model_cache/crazy-good-new-model-7b/fake_model-002-of-002.bin"
            in files_to_cache
        )

        assert "fake_model-001-of-002.bin" in model_files[model]["files"]
        assert "fake_model-001-of-002.bin" in model_files[model]["files"]


@patch("truss.contexts.image_builder.serving_image_builder.GCSCache.list_files")
def test_correct_nested_gcs_files_accessed_for_caching(mock_list_bucket_files):
    mock_list_bucket_files.return_value = [
        "folder_a/folder_b/fake_model-001-of-002.bin",
        "folder_a/folder_b/fake_model-002-of-002.bin",
    ]
    model = "gs://crazy-good-new-model-7b/folder_a/folder_b"

    config = TrussConfig(
        python_version="py39",
        model_cache=ModelCache(models=[ModelRepo(repo_id=model)]),
    )

    with TemporaryDirectory() as tmp_dir:
        truss_path = Path(tmp_dir)
        build_path = truss_path / "build"
        build_path.mkdir(parents=True, exist_ok=True)

        model_files, files_to_cache = get_files_to_cache(config, truss_path, build_path)
        files_to_cache = flatten_cached_files(files_to_cache)

        assert (
            "/app/model_cache/crazy-good-new-model-7b/folder_a/folder_b/fake_model-001-of-002.bin"
            in files_to_cache
        )
        assert (
            "/app/model_cache/crazy-good-new-model-7b/folder_a/folder_b/fake_model-002-of-002.bin"
            in files_to_cache
        )

        assert (
            "folder_a/folder_b/fake_model-001-of-002.bin" in model_files[model]["files"]
        )
        assert (
            "folder_a/folder_b/fake_model-001-of-002.bin" in model_files[model]["files"]
        )


@patch("truss.contexts.image_builder.serving_image_builder.S3Cache.list_files")
def test_correct_nested_s3_files_accessed_for_caching(mock_list_bucket_files):
    mock_list_bucket_files.return_value = [
        "folder_a/folder_b/fake_model-001-of-002.bin",
        "folder_a/folder_b/fake_model-002-of-002.bin",
    ]
    model = "s3://crazy-good-new-model-7b/folder_a/folder_b"

    config = TrussConfig(
        python_version="py39",
        model_cache=ModelCache(models=[ModelRepo(repo_id=model)]),
    )

    with TemporaryDirectory() as tmp_dir:
        truss_path = Path(tmp_dir)
        build_path = truss_path / "build"
        build_path.mkdir(parents=True, exist_ok=True)

        model_files, files_to_cache = get_files_to_cache(config, truss_path, build_path)
        files_to_cache = flatten_cached_files(files_to_cache)

        assert (
            "/app/model_cache/crazy-good-new-model-7b/folder_a/folder_b/fake_model-001-of-002.bin"
            in files_to_cache
        )
        assert (
            "/app/model_cache/crazy-good-new-model-7b/folder_a/folder_b/fake_model-002-of-002.bin"
            in files_to_cache
        )

        assert (
            "folder_a/folder_b/fake_model-001-of-002.bin" in model_files[model]["files"]
        )
        assert (
            "folder_a/folder_b/fake_model-001-of-002.bin" in model_files[model]["files"]
        )


@pytest.mark.integration
def test_truss_server_caching_truss():
    with ensure_kill_all():
        truss_root = (
            Path(__file__).parent.parent.parent.parent.parent.resolve() / "truss"
        )
        truss_dir = truss_root / "test_data" / "test_truss_server_caching_truss"
        tr = TrussHandle(truss_dir)

        container = tr.docker_run(
            local_port=8090, detach=True, wait_for_server_ready=True
        )
        time.sleep(15)
        assert "Downloading model.safetensors:" not in container.logs()


def test_model_cache_dockerfile():
    truss_root = Path(__file__).parent.parent.parent.parent.parent.resolve() / "truss"
    truss_dir = truss_root / "test_data" / "test_truss_server_caching_truss"
    tr = TrussHandle(truss_dir)

    builder_context = ServingImageBuilderContext
    image_builder = builder_context.run(tr.spec.truss_dir)

    secret_mount = f"RUN --mount=type=secret,id={HF_ACCESS_TOKEN_FILE_NAME}"
    with TemporaryDirectory() as tmp_dir:
        tmp_path = Path(tmp_dir)
        image_builder.prepare_image_build_dir(tmp_path, use_hf_secret=True)
        with open(tmp_path / "Dockerfile", "r") as f:
            gen_docker_file = f.read()
            assert secret_mount in gen_docker_file


def test_ignore_files_during_build_setup(custom_model_truss_dir_with_truss_ignore):
    th = TrussHandle(custom_model_truss_dir_with_truss_ignore)

    builder_context = ServingImageBuilderContext
    image_builder = builder_context.run(th.spec.truss_dir)

    ignore_files = ["random_file_1.txt"]
    ignore_folder = "random_folder_1/"
    do_not_ignore_folder = "random_folder_2/"

    with TemporaryDirectory() as build_dir:
        build_path = Path(build_dir)
        image_builder.prepare_image_build_dir(build_path)

        for file in ignore_files:
            assert not (build_path / file).exists()

        assert not (build_path / ignore_folder).exists()
        assert (build_path / do_not_ignore_folder).exists()


def test_trt_llm_build_dir(custom_model_trt_llm):
    th = TrussHandle(custom_model_trt_llm)
    builder_context = ServingImageBuilderContext
    image_builder = builder_context.run(th.spec.truss_dir)
    with TemporaryDirectory() as tmp_dir:
        tmp_path = Path(tmp_dir)
        image_builder.prepare_image_build_dir(tmp_path)
        build_th = TrussHandle(tmp_path)

        # Check that all files were copied
        _assert_copied(
            TRTLLM_TRUSS_DIR / "src",
            tmp_path / "server" / "extensions" / "trt_llm",
        )
        _assert_copied(
            TRTLLM_TRUSS_DIR / "packages",
            tmp_path / "packages",
        )

        assert (
            build_th.spec.config.runtime.predict_concurrency
            == TRTLLM_PREDICT_CONCURRENCY
        )
        assert build_th.spec.config.base_image.image == TRTLLM_BASE_IMAGE
        assert (
            build_th.spec.config.base_image.python_executable_path
            == TRTLLM_PYTHON_EXECUTABLE
        )
        assert BASE_TRTLLM_REQUIREMENTS == build_th.spec.config.requirements
        assert OPENAI_COMPATIBLE_TAG in build_th.spec.config.model_metadata["tags"]


def _assert_copied(src_path: str, dest_path: str):
    for dirpath, dirnames, filenames in os.walk(src_path):
        rel_path = os.path.relpath(dirpath, src_path)
        for filename in filenames:
            src_file = os.path.join(dirpath, filename)
            dest_file = os.path.join(dest_path, rel_path, filename)
            assert os.path.exists(dest_file), f"{dest_file} was not copied"
            assert filecmp.cmp(
                src_file, dest_file, shallow=False
            ), f"{src_file} and {dest_file} are not the same"



import pytest
from truss.errors import ValidationError
from truss.trt_llm.validation import _verify_has_class_init_arg, validate
from truss.truss_spec import TrussSpec


@pytest.mark.parametrize(
    "src, class_name, expected_arg, expected_to_raise",
    [
        (
            """
class Model:
    def __init__(self, foo):
        pass
            """,
            "Model",
            "foo",
            False,
        ),
        # Missing arg
        (
            """
class Model:
    def __init__(self, foo):
        pass
            """,
            "Model",
            "bar",
            True,
        ),
        # Missing class
        (
            """
class Model:
    def __init__(self, foo):
        pass
            """,
            "Missing",
            "bar",
            True,
        ),
        # With default value should still work
        (
            """
class Model:
    def __init__(self, foo=None):
        pass
            """,
            "Model",
            "foo",
            False,
        ),
        # not the only arg
        (
            """
class Model:
    def __init__(self, first, foo, third):
        pass
            """,
            "Model",
            "foo",
            False,
        ),
    ],
)
def test_has_class_init_arg(src, expected_arg, class_name, expected_to_raise):
    if expected_to_raise:
        with pytest.raises(ValidationError):
            _verify_has_class_init_arg(src, class_name, expected_arg)
    else:
        _verify_has_class_init_arg(src, class_name, expected_arg)


def test_validate(custom_model_trt_llm):
    spec = TrussSpec(custom_model_trt_llm)
    validate(spec)

    # overwrite with model code that doesn't take trt_llm as input
    override_model_code_invalid = """
class Model:
    def __init__(self):
        pass
"""
    spec.model_class_filepath.write_text(override_model_code_invalid)
    new_spec = TrussSpec(custom_model_trt_llm)
    with pytest.raises(ValidationError):
        validate(new_spec)

    # If model class file is removed, it should be ok
    spec.model_class_filepath.unlink()
    validate(new_spec)



import ast

from truss.errors import ValidationError
from truss.truss_spec import TrussSpec


def validate(truss_spec: TrussSpec):
    model_class_filepath = truss_spec.model_class_filepath
    if not model_class_filepath.exists():
        # It's ok if model class file is not provided,
        # trt_llm will generate one.
        return

    source = model_class_filepath.read_text()
    _verify_has_class_init_arg(source, truss_spec.model_class_name, "trt_llm")


def _verify_has_class_init_arg(source: str, class_name: str, arg_name: str):
    tree = ast.parse(source)
    model_class_init_found = False
    for node in tree.body:
        if isinstance(node, ast.ClassDef) and node.name == class_name:
            for child in node.body:
                if child.name == "__init__":  # type: ignore[attr-defined]
                    model_class_init_found = True
                    arg_found = False
                    for arg in child.args.args:  # type: ignore[attr-defined]
                        if arg.arg == arg_name:
                            arg_found = True
                    if not arg_found:
                        raise ValidationError(
                            (
                                "Model class `__init__` method is required to have `trt_llm` as an argument. Please add that argument.\n  "
                                "Or if you want to use the automatically generated model class then remove the `model.py` file."
                            )
                        )

    if not model_class_init_found:
        raise ValidationError(
            "Model class does not have an `__init__` method; when using `trt_llm` it is required"
        )



import rich
from InquirerPy import inquirer


def inquire_copy_confirm(msg: str) -> bool:
    rich.print(msg)
    return inquirer.confirm(
        "Do you want to proceed with deploying and copying "
        "all data (⚠️ proceed with cation)?",
        default=False,
    ).execute()



from pathlib import Path
from typing import Dict, List, Optional

from truss.patch.hash import file_content_hash_str
from truss.util.path import get_unignored_relative_paths_from_root


def directory_content_signature(
    root: Path, ignore_patterns: Optional[List[str]] = None
) -> Dict[str, str]:
    """Calculate content signature of a filesystem directory.

    Sort all files by path, store file path with content hash.
    Signatures are meant to track changes in directory. e.g.
    A previous signature and the directory can be combined to create
    a patch from the previous state.

    Hash of directories is marked None.
    """
    paths = list(get_unignored_relative_paths_from_root(root, ignore_patterns))
    paths.sort()

    def path_hash(pth: Path):
        if pth.is_file():
            return file_content_hash_str(pth)

    return {str(path): path_hash(root / path) for path in paths}



import collections
import inspect
import logging
import pathlib
import tempfile
import uuid
from typing import (
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    MutableMapping,
    Optional,
    Type,
    cast,
)

import truss
from tenacity import retry, stop_after_delay, wait_fixed
from truss.remote.baseten import remote as b10_remote
from truss.remote.baseten import service as b10_service
from truss.remote.baseten import types as b10_types

from truss_chains import code_gen, definitions, framework, utils


def _deploy_to_baseten(
    truss_dir: pathlib.Path, options: definitions.DeploymentOptionsBaseten
) -> b10_service.BasetenService:
    truss_handle = truss.load(str(truss_dir))
    model_name = truss_handle.spec.config.model_name
    assert model_name is not None
    if options.promote and not options.publish:
        logging.info("`promote=True` overrides `publish` to `True`.")
    logging.info(
        f"Deploying chainlet `{model_name}` as a truss model on Baseten "
        f"(publish={options.publish}, promote={options.promote})."
    )

    # Since we are deploying a model independently of the chain, we add a random suffix to
    # prevent us from running into issues with existing models with the same name.
    #
    # This is a bit of a hack for now. Once we support model_origin for Chains models, we
    # can drop the requirement for names on models.
    model_suffix = str(uuid.uuid4()).split("-")[0]

    # Models must be trusted to use the API KEY secret.
    service = options.remote_provider.push(
        truss_handle,
        model_name=f"{model_name}-{model_suffix}",
        trusted=True,
        publish=options.publish,
        promote=options.promote,
        origin=b10_types.ModelOrigin.CHAINS,
    )
    return cast(b10_service.BasetenService, service)


class DockerService(b10_service.TrussService):
    def __init__(self, remote_url: str, is_draft: bool, **kwargs):
        super().__init__(remote_url, is_draft, **kwargs)

    def authenticate(self) -> Dict[str, str]:
        return {}

    def is_live(self) -> bool:
        response = self._send_request(self._service_url, "GET")
        if response.status_code == 200:
            return True
        return False

    def is_ready(self) -> bool:
        response = self._send_request(self._service_url, "GET")
        if response.status_code == 200:
            return True
        return False

    @property
    def logs_url(self) -> str:
        raise NotImplementedError()

    @property
    def predict_url(self) -> str:
        return f"{self._service_url}/v1/models/model:predict"

    def poll_deployment_status(self, sleep_secs: int = 1) -> Iterator[str]:
        raise NotImplementedError()


def _deploy_service(
    truss_dir: pathlib.Path,
    chainlet_descriptor: definitions.ChainletAPIDescriptor,
    options: definitions.DeploymentOptions,
) -> Optional[b10_service.TrussService]:
    service: Optional[b10_service.TrussService]
    if options.only_generate_trusses:
        service = None
    elif isinstance(options, definitions.DeploymentOptionsLocalDocker):
        logging.info(f"Running in docker container `{chainlet_descriptor.name}` ")
        port = utils.get_free_port()
        truss_handle = truss.load(str(truss_dir))
        truss_handle.add_secret(
            definitions.BASETEN_API_SECRET_NAME, options.baseten_chain_api_key
        )
        truss_handle.docker_run(
            local_port=port,
            detach=True,
            wait_for_server_ready=True,
            network="host",
            container_name_prefix=chainlet_descriptor.name,
        )
        # http://localhost:{port} seems to only work *sometimes* with docker.
        service = DockerService(f"http://host.docker.internal:{port}", is_draft=True)
    elif isinstance(options, definitions.DeploymentOptionsBaseten):
        with utils.log_level(logging.INFO):
            service = _deploy_to_baseten(truss_dir, options)
    else:
        raise NotImplementedError(options)

    if service:
        logging.info(
            f"Service created for `{chainlet_descriptor.name}` @ {service.predict_url}."
        )
    return service


def _get_ordered_dependencies(
    chainlets: Iterable[Type[definitions.ABCChainlet]],
) -> Iterable[definitions.ChainletAPIDescriptor]:
    """Gather all Chainlets needed and returns a topologically ordered list."""
    needed_chainlets: set[definitions.ChainletAPIDescriptor] = set()

    def add_needed_chainlets(chainlet: definitions.ChainletAPIDescriptor):
        needed_chainlets.add(chainlet)
        for chainlet_descriptor in framework.global_chainlet_registry.get_dependencies(
            chainlet
        ):
            needed_chainlets.add(chainlet_descriptor)
            add_needed_chainlets(chainlet_descriptor)

    for chainlet_cls in chainlets:
        add_needed_chainlets(
            framework.global_chainlet_registry.get_descriptor(chainlet_cls)
        )
    # Iterating over the registry ensures topological ordering.
    return [
        descr
        for descr in framework.global_chainlet_registry.chainlet_descriptors
        if descr in needed_chainlets
    ]


def _chainlet_logs_url(
    chain_id: str,
    chain_deployment_id: str,
    chainlet_id: str,
    remote: b10_remote.BasetenRemote,
) -> str:
    return f"{remote._remote_url}/chains/{chain_id}/logs/{chain_deployment_id}/{chainlet_id}"


def _chain_status_page_url(
    chain_id: str,
    remote: b10_remote.BasetenRemote,
) -> str:
    return f"{remote._remote_url}/chains/{chain_id}/overview"


class RemoteChainService:
    _remote: (
        b10_remote.BasetenRemote
    )  # TODO, make this a generic TypeVar for this calss
    _chain_id: str
    _chain_deployment_id: str

    def __init__(
        self, chain_id: str, chain_deployment_id: str, remote: b10_remote.BasetenRemote
    ) -> None:
        self._remote = remote
        self._chain_id = chain_id
        self._chain_deployment_id = chain_deployment_id

    @retry(stop=stop_after_delay(300), wait=wait_fixed(1), reraise=True)
    def get_info(self) -> list[b10_types.DeployedChainlet]:
        """Return list with elements (name, status, logs_url) for each chainlet."""
        chainlets = self._remote.get_chainlets(
            chain_deployment_id=self._chain_deployment_id
        )

        return [
            b10_types.DeployedChainlet(
                name=chainlet["name"],
                is_entrypoint=chainlet["is_entrypoint"],
                status=chainlet["oracle_version"]["current_model_deployment_status"][
                    "status"
                ],
                logs_url=_chainlet_logs_url(
                    self._chain_id,
                    self._chain_deployment_id,
                    chainlet["id"],
                    self._remote,
                ),
            )
            for chainlet in chainlets
        ]

    @property
    def status_page_url(self) -> str:
        return _chain_status_page_url(self._chain_id, self._remote)


class ChainService:
    # TODO: this exposes methods to users that should be internal (e.g. `add_service`).
    """Handle for a deployed chain.

    A ``ChainService`` is created and returned when using ``deploy_remotely``. It
    bundles the individual services for each chainlet in the chain, and provides
    utilities to query their status, invoke the entrypoint etc.
    """

    name: str
    _entrypoint: str
    _services: MutableMapping[str, b10_service.TrussService]
    _entrypoint_fake_json_data = Any
    _remote_chain_service: Optional[RemoteChainService]

    def __init__(self, entrypoint: str, name: str) -> None:
        """
        Args:
            entrypoint: Name of the entrypoint chainlet.
            name: Name of the chain.
        """
        self.name = name
        self._entrypoint = entrypoint
        self._services = collections.OrderedDict()  # Preserve order.
        self.entrypoint_fake_json_data = None

    def add_service(self, name: str, service: b10_service.TrussService) -> None:
        """
        Used to add a chainlet service during the deployment sequence of the chain.


        Args:
            name: Chainlet name.
            service: Service object for the chainlet.
        """
        self._services[name] = service

    def set_remote_chain_service(
        self, chain_id: str, chain_deployment_id: str, remote: b10_remote.BasetenRemote
    ) -> None:
        self._remote_chain_service = RemoteChainService(
            chain_id, chain_deployment_id, remote
        )

    @property
    def entrypoint_fake_json_data(self) -> Any:
        """Fake JSON example data that matches the entrypoint's input schema.
        This property must be externally populated.

        Raises:
            ValueError: If fake data was not set.
        """
        if self._entrypoint_fake_json_data is None:
            raise ValueError("Fake data was not set.")
        return self._entrypoint_fake_json_data

    @entrypoint_fake_json_data.setter
    def entrypoint_fake_json_data(self, fake_data: Any) -> None:
        self._entrypoint_fake_json_data = fake_data

    @property
    def get_entrypoint(self) -> b10_service.TrussService:
        """Returns the entrypoint's service handle.

        Raises:
            MissingDependencyError: If the entrypoint service was not added.
        """
        service = self._services.get(self._entrypoint)
        if not service:
            raise definitions.MissingDependencyError(
                f"Service for entrypoint `{self._entrypoint}` was not added."
            )
        return service

    @property
    def services(self) -> MutableMapping[str, b10_service.TrussService]:
        return self._services

    @property
    def entrypoint_name(self) -> str:
        return self._entrypoint

    @property
    def run_url(self) -> str:
        """URL to invoke the entrypoint."""
        return self.get_entrypoint.predict_url

    def run_remote(self, json: Dict) -> Any:
        """Invokes the entrypoint with JSON data.

        Returns:
            The JSON response."""
        return self.get_entrypoint.predict(json)

    def get_info(self) -> list[b10_types.DeployedChainlet]:
        """Queries the statuses of all chainlets in the chain.

        Returns:
            List with elements ``(name, status, logs_url)`` for each chainlet."""
        if not self._remote_chain_service:
            raise ValueError("Chain was not deployed remotely.")

        return self._remote_chain_service.get_info()

    @property
    def status_page_url(self) -> str:
        """Link to status page on Baseten."""
        if not self._remote_chain_service:
            raise ValueError("Chain was not deployed remotely.")

        return self._remote_chain_service.status_page_url


def deploy_remotely(
    entrypoint: Type[definitions.ABCChainlet],
    options: definitions.DeploymentOptions,
    non_entrypoint_root_dir: Optional[str] = None,
    gen_root: pathlib.Path = pathlib.Path(tempfile.gettempdir()),
) -> ChainService:
    # TODO: revisit how chain root is inferred/specified, current might be brittle.
    if non_entrypoint_root_dir:
        chain_root = pathlib.Path(non_entrypoint_root_dir).absolute()
    else:
        chain_root = pathlib.Path(inspect.getfile(entrypoint)).absolute().parent
    logging.info(
        f"Using chain workspace dir: `{chain_root}` (files under this dir will "
        "be included as dependencies in the remote deployments and are importable)."
    )

    chainlet_name_to_url: dict[str, str] = {}
    chain_service = ChainService(
        framework.global_chainlet_registry.get_descriptor(entrypoint).name,
        name=options.chain_name,
    )
    if isinstance(options, definitions.DeploymentOptionsBaseten):
        secrets_info = options.remote_provider.api.get_all_secrets()
        secret_names = {sec["name"] for sec in secrets_info["secrets"]}
        if definitions.BASETEN_API_SECRET_NAME not in secret_names:
            logging.info(
                "It seems you are using chains for the first time, since there "
                f"is no `{definitions.BASETEN_API_SECRET_NAME}` secret on baseten. "
                "Creating secret automatically."
            )
            options.remote_provider.api.upsert_secret(
                definitions.BASETEN_API_SECRET_NAME,
                options.remote_provider.api.auth_token.value,
            )

    for chainlet_descriptor in _get_ordered_dependencies([entrypoint]):
        deps = framework.global_chainlet_registry.get_dependencies(chainlet_descriptor)
        chainlet_dir = code_gen.gen_truss_chainlet(
            options,
            chainlet_descriptor,
            deps,
            chainlet_name_to_url,
            chain_root,
            gen_root,
        )
        service = _deploy_service(chainlet_dir, chainlet_descriptor, options)
        if service:
            chain_service.add_service(chainlet_descriptor.name, service)
            chainlet_name_to_url[chainlet_descriptor.name] = service.predict_url
        else:  # only_generate_trusses case.
            chainlet_name_to_url[chainlet_descriptor.name] = "http://dummy"

    if (
        isinstance(options, definitions.DeploymentOptionsBaseten)
        and not options.only_generate_trusses
    ):
        chainlets: List[b10_types.ChainletData] = []
        entrypoint_name = chain_service.entrypoint_name

        for chainlet_name, truss_service in chain_service.services.items():
            baseten_service = cast(b10_service.BasetenService, truss_service)
            chainlets.append(
                b10_types.ChainletData(
                    name=chainlet_name,
                    oracle_version_id=baseten_service.model_version_id,
                    is_entrypoint=chainlet_name == entrypoint_name,
                )
            )

        chain_deployment_handle = options.remote_provider.create_chain(
            chain_name=chain_service.name,
            chainlets=chainlets,
            publish=options.publish,
            promote=options.promote,
        )

        chain_service.set_remote_chain_service(
            chain_deployment_handle.chain_id,
            chain_deployment_handle.chain_deployment_id,
            options.remote_provider,
        )

    return chain_service



import click
import pytest
import requests_mock
from truss.remote.baseten.core import ModelId, ModelName, ModelVersionId
from truss.remote.baseten.remote import BasetenRemote
from truss.truss_handle import TrussHandle

_TEST_REMOTE_URL = "http://test_remote.com"
_TEST_REMOTE_GRAPHQL_PATH = "http://test_remote.com/graphql/"


def test_get_service_by_version_id():
    remote = BasetenRemote(_TEST_REMOTE_URL, "api_key")

    version = {
        "id": "version_id",
        "oracle": {
            "id": "model_id",
        },
    }
    model_version_response = {"data": {"model_version": version}}

    with requests_mock.Mocker() as m:
        m.post(
            _TEST_REMOTE_GRAPHQL_PATH,
            json=model_version_response,
        )
        service = remote.get_service(model_identifier=ModelVersionId("version_id"))

    assert service.model_id == "model_id"
    assert service.model_version_id == "version_id"


def test_get_service_by_version_id_no_version():
    remote = BasetenRemote(_TEST_REMOTE_URL, "api_key")
    model_version_response = {"errors": [{"message": "error"}]}
    with requests_mock.Mocker() as m:
        m.post(
            _TEST_REMOTE_GRAPHQL_PATH,
            json=model_version_response,
        )
        with pytest.raises(click.UsageError):
            remote.get_service(model_identifier=ModelVersionId("version_id"))


def test_get_service_by_model_name():
    remote = BasetenRemote(_TEST_REMOTE_URL, "api_key")

    versions = [
        {"id": "1", "is_draft": False, "is_primary": False},
        {"id": "2", "is_draft": False, "is_primary": True},
        {"id": "3", "is_draft": True, "is_primary": False},
    ]
    model_response = {
        "data": {
            "model": {
                "name": "model_name",
                "id": "model_id",
                "versions": versions,
            }
        }
    }

    with requests_mock.Mocker() as m:
        m.post(
            _TEST_REMOTE_GRAPHQL_PATH,
            json=model_response,
        )

        # Check that the production version is returned when published is True.
        service = remote.get_service(
            model_identifier=ModelName("model_name"), published=True
        )
        assert service.model_id == "model_id"
        assert service.model_version_id == "2"

        # Check that the development version is returned when published is False.
        service = remote.get_service(
            model_identifier=ModelName("model_name"), published=False
        )
        assert service.model_id == "model_id"
        assert service.model_version_id == "3"


def test_get_service_by_model_name_no_dev_version():
    remote = BasetenRemote(_TEST_REMOTE_URL, "api_key")

    versions = [
        {"id": "1", "is_draft": False, "is_primary": True},
    ]
    model_response = {
        "data": {
            "model": {
                "name": "model_name",
                "id": "model_id",
                "versions": versions,
            }
        }
    }

    with requests_mock.Mocker() as m:
        m.post(
            _TEST_REMOTE_GRAPHQL_PATH,
            json=model_response,
        )

        # Check that the production version is returned when published is True.
        service = remote.get_service(
            model_identifier=ModelName("model_name"), published=True
        )
        assert service.model_id == "model_id"
        assert service.model_version_id == "1"

        # Since no development version exists, calling get_service with
        # published=False should raise an error.
        with pytest.raises(click.UsageError):
            remote.get_service(
                model_identifier=ModelName("model_name"), published=False
            )


def test_get_service_by_model_name_no_prod_version():
    remote = BasetenRemote(_TEST_REMOTE_URL, "api_key")

    versions = [
        {"id": "1", "is_draft": True, "is_primary": False},
    ]
    model_response = {
        "data": {
            "model": {
                "name": "model_name",
                "id": "model_id",
                "versions": versions,
            }
        }
    }

    with requests_mock.Mocker() as m:
        m.post(
            _TEST_REMOTE_GRAPHQL_PATH,
            json=model_response,
        )

        # Since no production version exists, calling get_service with
        # published=True should raise an error.
        with pytest.raises(click.UsageError):
            remote.get_service(model_identifier=ModelName("model_name"), published=True)

        # Check that the development version is returned when published is False.
        service = remote.get_service(
            model_identifier=ModelName("model_name"), published=False
        )
        assert service.model_id == "model_id"
        assert service.model_version_id == "1"


def test_get_service_by_model_id():
    remote = BasetenRemote(_TEST_REMOTE_URL, "api_key")

    model_response = {
        "data": {
            "model": {
                "name": "model_name",
                "id": "model_id",
                "primary_version": {"id": "version_id"},
            }
        }
    }

    with requests_mock.Mocker() as m:
        m.post(
            _TEST_REMOTE_GRAPHQL_PATH,
            json=model_response,
        )

        service = remote.get_service(model_identifier=ModelId("model_id"))
        assert service.model_id == "model_id"
        assert service.model_version_id == "version_id"


def test_get_service_by_model_id_no_model():
    remote = BasetenRemote(_TEST_REMOTE_URL, "api_key")
    model_response = {"errors": [{"message": "error"}]}
    with requests_mock.Mocker() as m:
        m.post(
            _TEST_REMOTE_GRAPHQL_PATH,
            json=model_response,
        )
        with pytest.raises(click.UsageError):
            remote.get_service(model_identifier=ModelId("model_id"))


def test_push_raised_value_error_when_deployment_name_and_not_publish(
    custom_model_truss_dir_with_pre_and_post,
):
    remote = BasetenRemote(_TEST_REMOTE_URL, "api_key")
    model_response = {
        "data": {
            "model": {
                "name": "model_name",
                "id": "model_id",
                "primary_version": {"id": "version_id"},
            }
        }
    }
    with requests_mock.Mocker() as m:
        m.post(
            _TEST_REMOTE_GRAPHQL_PATH,
            json=model_response,
        )
        th = TrussHandle(custom_model_truss_dir_with_pre_and_post)

        with pytest.raises(
            ValueError,
            match="Deployment name cannot be used for development deployment",
        ):
            remote.push(th, "model_name", False, False, False, False, "dep_name")


def test_push_raised_value_error_when_deployment_name_is_not_valid(
    custom_model_truss_dir_with_pre_and_post,
):
    remote = BasetenRemote(_TEST_REMOTE_URL, "api_key")
    model_response = {
        "data": {
            "model": {
                "name": "model_name",
                "id": "model_id",
                "primary_version": {"id": "version_id"},
            }
        }
    }
    with requests_mock.Mocker() as m:
        m.post(
            _TEST_REMOTE_GRAPHQL_PATH,
            json=model_response,
        )
        th = TrussHandle(custom_model_truss_dir_with_pre_and_post)

        with pytest.raises(
            ValueError,
            match="Deployment name must only contain alphanumeric, -, _ and . characters",
        ):
            remote.push(th, "model_name", True, False, False, False, "dep//name")


def test_push_raised_value_error_when_keep_previous_prod_settings_and_not_promote(
    custom_model_truss_dir_with_pre_and_post,
):
    remote = BasetenRemote(_TEST_REMOTE_URL, "api_key")
    model_response = {
        "data": {
            "model": {
                "name": "model_name",
                "id": "model_id",
                "primary_version": {"id": "version_id"},
            }
        }
    }
    with requests_mock.Mocker() as m:
        m.post(
            _TEST_REMOTE_GRAPHQL_PATH,
            json=model_response,
        )
        th = TrussHandle(custom_model_truss_dir_with_pre_and_post)

        with pytest.raises(
            ValueError,
            match="preserve-previous-production-deployment can only be used with the '--promote' option",
        ):
            remote.push(th, "model_name", False, False, False, True)



from typing import Any


class Model:
    def __init__(self, trt_llm, **kwargs) -> None:
        self._data_dir = kwargs["data_dir"]
        self._config = kwargs["config"]
        self._secrets = kwargs["secrets"]
        self._model = None
        self._engine = trt_llm["engine"]

    async def predict(self, model_input: Any) -> Any:
        return await self._engine.predict(model_input)
